# Generated by roxygen2: do not edit by hand

S3method(gesearch,default)
S3method(gesearch,formula)
S3method(plot,gesearch)
S3method(plot,liblex)
S3method(plot,mbl)
S3method(plot,ortho_projection)
S3method(predict,gesearch)
S3method(predict,liblex)
S3method(predict,ortho_projection)
S3method(predict,plsr)
S3method(predict,resemble_model)
S3method(print,diss_correlation)
S3method(print,diss_method)
S3method(print,diss_pca)
S3method(print,diss_pls)
S3method(print,fit_gpr)
S3method(print,fit_method)
S3method(print,fit_pls)
S3method(print,fit_wapls)
S3method(print,gesearch)
S3method(print,liblex)
S3method(print,mbl)
S3method(print,model_control)
S3method(print,ncomp_selection)
S3method(print,neighbors)
S3method(print,neighbors_diss)
S3method(print,neighbors_k)
S3method(print,ortho_projection)
S3method(print,resemble_dissimilarity)
S3method(print,resemble_model)
export(diss_correlation)
export(diss_cosine)
export(diss_euclidean)
export(diss_evaluate)
export(diss_mahalanobis)
export(diss_pca)
export(diss_pls)
export(dissimilarity)
export(fit_gpr)
export(fit_pls)
export(fit_wapls)
export(gesearch)
export(gesearch_control)
export(get_predictions)
export(liblex)
export(liblex_control)
export(mbl)
export(mbl_control)
export(model)
export(model_control)
export(ncomp_by_cumvar)
export(ncomp_by_opc)
export(ncomp_by_var)
export(ncomp_fixed)
export(neighbors_diss)
export(neighbors_k)
export(ortho_projection)
export(search_neighbors)
export(sid)
export(sim_eval)
import(Rcpp)
import(foreach)
import(grDevices)
import(graphics)
import(iterators)
import(lifecycle)
import(mathjaxr)
importFrom(RhpcBLASctl,blas_get_num_procs)
importFrom(RhpcBLASctl,blas_set_num_threads)
importFrom(foreach,"%dopar%")
importFrom(foreach,foreach)
importFrom(graphics,barplot)
importFrom(graphics,grid)
importFrom(graphics,par)
importFrom(graphics,plot)
importFrom(graphics,segments)
importFrom(lifecycle,deprecate_soft)
importFrom(mathjaxr,preview_rd)
importFrom(stats,.MFclass)
importFrom(stats,approx)
importFrom(stats,as.formula)
importFrom(stats,bw.nrd0)
importFrom(stats,complete.cases)
importFrom(stats,cor)
importFrom(stats,cov)
importFrom(stats,delete.response)
importFrom(stats,density)
importFrom(stats,diffinv)
importFrom(stats,median)
importFrom(stats,model.extract)
importFrom(stats,model.frame)
importFrom(stats,model.matrix)
importFrom(stats,na.fail)
importFrom(stats,na.omit)
importFrom(stats,na.pass)
importFrom(stats,predict)
importFrom(stats,quantile)
importFrom(stats,reshape)
importFrom(stats,sd)
importFrom(stats,setNames)
importFrom(stats,terms)
importFrom(stats,update)
importFrom(utils,flush.console)
importFrom(utils,setTxtProgressBar)
importFrom(utils,tail)
importFrom(utils,txtProgressBar)
useDynLib(resemble, .registration=TRUE)
